UNIT MMCLoader16_32;
(*
   The MMCLoader16_32 handles the MMC / SD / SDHC access, Filereading and the Flashing for the CPULoader. It detects and supports FAT16 and FAT32
*)

INTERFACE

{ $W+}                                                                          // enable/disable warnings for this unit

TYPE
  tBootFlag          = Record
                          MinPrevVersion : WORD;                                // Minimum previous version required for the Bootloader to start up and flash
                          Version        : WORD;                                // The version of this file (Build-No)
                          Flag           : WORD;
                       End;

CONST
  Application_Start         : LONGWORD  = $00000;                               // Pointer to the first Byte of the Application Area. Here we will start flashing. Should always be 0.
{!}
  Boot_Start                : LONGWORD  = $3F000;                               // Pointer to the first Byte of the CPULoader-Area: must be adjusted depending your hardware and your FuseBits...
                                                                                // This is a BYTE-address!! (not a WORD address)
  BootFlag[@Boot_Start - 8] : tBootFlag = (MinPrevVersion : $FFFF;              // The minimal required previous Build No. of the APPLICATION - not of the CPULoader!
                                           Version        : $FFFF;              // Do not change these values set here.
                                           Flag           : $FFFF);             // After the BootFlag.Flag another word follows with the checksum of the complete file at [@BOOT_START - 2]
                                                                                // These 3 variables (Version, Flag, CheckSum) must be in the binary file of the application and
                                                                                // thus must be defined there!! They are the last 8 bytes of the APPLICATION memory.
{!}
  Boot_End                  : LongWord  = $40000;                               // Pointer to the first Byte AFTER the End of the Boot Area
  LoaderFlag[@Boot_End - 8] : tBootFlag = (MinPrevVersion : $0000;              // The minimal required previous Build No. of the CPULoader
                                           Version        : PROJECTBUILD+1;     // BuildNo of the BootLoader
                                           Flag           : $BB66);             // Word-Flag, which contains a marker (here $BB66) for the CPULoader.

                                                                                // After the LoaderFlag.Flag another word follows with the checksum of the complete file at [@BOOT_END - 2]
                                                                                // These 4 variables (MinPrevVersion, Version, Flag, CheckSum) must be in the binary file of the application and
                                                                                // thus must be defined there!! They are the last 8 bytes of the Bootloader memory.
                                                                                // ALWAYS check, that the code does NOT overwrite this constant (it is a close shave!)


{!} // -->
  UPDNAME                 : STRING  = 'CPU     BIN';                            // Filename is CPU.BIN - This is the name as it is compared here. So, here the extension is used, note without the "." in between
{!} // <--

{$IDATA}

VAR                                                                             // Public variables, procedures and functions of this unit..
{-global variables -----------------------------------------------------------}
   gMinPrevVersion : Word;
   gFileVersion    : Word;
   gFileCheckSum   : Word;


PROCEDURE UpdateFirmware;                                                       // This reads the UpdateFile and flashes it into the application area.
FUNCTION  FATInit                   : BOOLEAN;                                  // Initialize the MMC and check for a valid FAT16 file system


IMPLEMENTATION

CONST
  SPIINITCLOCK            : BYTE = %01010011;	                                // f/128
  SPIREADCLOCK            : BYTE = %01010000;	                                // f/4
  SPIWRITECLOCK           : BYTE = %01010010;	                                // f/64

                                                                                //* Definitions for MMC/SDC command */
  CMD0           : byte        = ($40 + 0);                                     // GO_IDLE_STATE
  CMD1           : byte        = ($40 + 1);                                     // SEND_OP_COND (MMC)
  ACMD41         : byte        = ($C0 + 41);                                    // SEND_OP_COND (SDC)
  CMD8           : byte        = ($40 + 8);                                     // SEND_IF_COND
  CMD16          : byte        = ($40 + 16);                                    // SET_BLOCKLEN
  CMD17          : byte        = ($40 + 17);                                    // Read Sector
  CMD55          : byte        = ($40 + 55);                                    // APP_CMD
  CMD58          : byte        = ($40 + 58);                                    // READ_OCR

                                                                                //* Card type flags (CardType) */
  CT_MMC   : BYTE = $01;                                                        // Old MMC Card
  CT_SD1   : BYTE = $02;                                                        // SD Card Version 1
  CT_SD2   : BYTE = $04;                                                        // SD Card Version 2
  CT_BLOCK : BYTE = $08;                                                        // together with CT_SD2 it indicates a SDHC card



TYPE
  tFATtype       = (tFATnone, tFAT16, tFAT32);                                  // FAT System types. tFATnone = 0; tFAT16 = 1, tFAT32 = 2 --> DO NOT CHANGE THIS!!
  
  tDirEntry   = RECORD                                                          // File/Directory Entry in the FAT 16 / 32
                  Name_Ext  : ARRAY[1..11] OF CHAR;                             // Name of the File together with the Extension
                  Attrib    : BYTE;
                  res       : ARRAY[1..8] OF BYTE;
                  ClusterHi : WORD;                                             // HiWord of the first Cluster of the File (FAT32 only, always 0 with FAT16)
                  Time      : WORD;
                  Date      : WORD;
                  ClusterLo : WORD;                                             // LoWord of the first Cluster of the File (FAT 16 / 32)
                  Size      : LONGWORD;                                         // Size of the File, for Directories always 0
                END;
                
  tDirPtr     = POINTER TO tDirEntry;
  tDirEntries = ARRAY[1..16] OF tDirEntry;

VAR
{$PDATA}
{!}
  MMC_CS  [@PortB, 0] : BIT;                                                    // Check, if this is your hardware SPI Port that is connected to the MMC / SD / SDHC

  SPIF    [@SPSR, 7]  : Bit;
  
{$IDATA}
{-Buffer und Hilfsvariablen----------------------------------------------------}
  SDDummy           : BYTE;
  buffer            : ARRAY[0..511] OF BYTE;                                    // This BYTE buffer is filled from the Update File
  flash[@buffer]    : ARRAY[0..255] OF WORD;                                    // This is the very same buffer, just addressed as WORDs
  Dir[@buffer]      : tDirEntries;                                              // This is the very same buffer, just addressed as Directory Entries
  WordsFlashed      : LONGWORD;                                                 // Counter for the WORDS flashed by the update routine
  CardType          : Byte;                                                     // 0 = no Card, 1 = MMC-Card, 2 = SD V1-Card, 4 = SD V2-Card, 12 = SDHC-Card

{-PTI Partition Info-----------------------------------------------------------}
  PTI_StartSector   : LONGWORD;                                                 // Offset of the first Sector of the FAT System. Get this first from the MasterBootRecord

CONST
{-BPB BIOS Parameter Block-----------------------------------------------------}
  BPB_BytesPerSec     : WORD = 512;                                             // Number of Bytes per Sector - this is ALWAYS 512 Bytes, as recommended by Microsoft

VAR
{-BPB BIOS Parameter Block-----------------------------------------------------}
  BPB_SecPerCluster : BYTE;                                                     // How many sectors make one Cluster: this is set when formatting the Card
  BPB_RsvdSecCnt    : WORD;                                                     // Number of rewerved sectors after the first Sector of the partition
  BPB_NumFATs       : BYTE;                                                     // Number of FAT tables on the partition, normally 2
  BPB_FatSz         : LONGWORD;                                                 // Number of Sectors on ONE FAT

{-FAT--------------------------------------------------------------------------}
  FAT_RootDir       : LONGWORD;                                                 // First sector of the FAT16 Root Dir. For FAT32 this is the Start of the FAT (= FAT_StartData), because FAT32 has no separate Root Dir.
  FAT_StartData     : LONGWORD;                                                 // First sector of the FAT 16 / 32 Data
  FATver            : tFATtype;                                                 // Type of FAT System: supported are FAT 16 / 32
  FAT_RootDirSecs   : WORD;                                                     // Number of Sectors for the RootDir - FAT16 only, for FAT32 this is 0

{-FILE-------------------------------------------------------------------------}
  FirstCluster      : LONGWORD;                                                 // Variable, that points to the First Cluster of a File
  NextCluster       : LONGWORD;                                                 // Variable, that points to the Next  Cluster of a File



{-FUNCTIONS AND PROCEDURES-----------------------------------------------------}

{===============================================================================
  SPIPut
===============================================================================}
PROCEDURE SPIPut(b: Byte);                                                      // Write one Byte to the Card via SPI
BEGIN
  SPDR := b;
  REPEAT UNTIL SPIF;
END SPIPut;


{===============================================================================
  SPIGet
===============================================================================}
FUNCTION SPIGet : BYTE;                                                         // Read one Byte from the Card via SPI
BEGIN
  SPDR := $FF;
  REPEAT UNTIL SPIF;
  RETURN (SPDR);
END SPIGet;


{===============================================================================
  MMCSendCmd
===============================================================================}
FUNCTION MMCSendCmd(cmd, argA, argB, argC, argD, CRC : BYTE) : BYTE;            // Send a Command to the Card via SPI
VAR
  result  : BYTE;
  i       : WORD;
BEGIN
  If ((cmd AND $80) <> 0) Then                                                  // If the cmd is a ACMD (Most Significant Bit is Set) = Application Specific Command
    result := MMCSendCmd(CMD55, 0, 0, 0, 0, $FF);                               // it MUST be preceded by a CMD55 to announce it
    cmd := cmd AND $7F;                                                         // and the MSB must be deleted
    If (result > 1) Then
      RETURN (result);
    EndIf;
  EndIf;

  SPIPut($FF);                                                                  // Delay
  EXCL(MMC_CS);                                                                 // Activate the CS signal (Lo = active)
  SPIPut(cmd);                                                                  // And write the Command with the CRC to the Card
  SPIPut(argA);
  SPIPut(argB);
  SPIPut(argC);
  SPIPut(argD);
  SPIPut(CRC);
  i := 1;
  WHILE i <= $1FFF DO                                                           // Wait for the Card to send the Result
    result := SPIGet;                                                           // Get the Result
    IF ((result and $80) = 0) THEN
      Break;
    ENDIF;
    Inc(i);
  ENDWHILE;
  RETURN (result);                                                              // Return the Result
END MMCSendCmd;


{===============================================================================
  CARD_INIT
===============================================================================}
Procedure CardInit;                                                             //  Initializes the SD-Card or MMC and reads in the Card Type
Var
  n, res  : BYTE;
  cmd : BYTE;
  ocr : Array[0..3] of BYTE;
  i, tmr : WORD;
Begin
  CardType := 0;
  INCL(MMC_CS);                                                                 // Set the Chip Select (CS) Signal to active (Lo)
{!} // -->                                                                      //  You may have to change the DDRs of Port B here to your processors hardware SPI Port...
  INCL(DDRB, 0);  //CS
  INCL(DDRB, 1);  //SCK ist output
  INCL(DDRB, 2);  //MOSI ist output
  EXCL(DDRB, 0);
  EXCL(DDRB, 1);
  EXCL(DDRB, 2);
  INCL(DDRB, 0);  //CS ist output
  INCL(DDRB, 1);  //SCK ist output
  INCL(DDRB, 2);  //MOSI ist output
  EXCL(DDRB, 3);  //MISO
{!} // <--

  SPCR := SPIINITCLOCK;                                                         // Set the Speed of the SPI bus
  INCL(SPSR, 0);
  i := 1;
  WHILE i <= 10 DO
    SPIGet;
    Inc(i);
  ENDWHILE;
  {---------------------------------------------}
  i := 1;
  WHILE i <= 512 DO                                                             // Try up to 512 times to RESET the MMC
    res := MMCSendCmd(CMD0, 0, 0, 0, 0, $95);                                   // Reset = CMD0. Here, the correct CRC ($95) must be sent, so we calculate it manually and put it here
    INCL(MMC_CS);
    SPIGet;                                                                     // Read back the Result of the RESET from the MMC
    IF res = $01 THEN                                                           // Result was OK, so leave the loop.
      BREAK;
    ENDIF;
    Inc(i);
  ENDWHILE;

  For n := 100 DownTo 1 Do
    SPIGet;
  EndFor;

    If (MMCSendCmd(CMD8, 0, 0, $1, $AA, $87) = 1) Then                          // Check the Card's interface condition. If the Answer is 1, the card is SD Version 2 or SDHC and supports 2.7 V - 3.3 V
      For n := 0 To 3 Do                                                        // Also with CMD8 the CRC is mandatory, so we have added the correct value ($87) for this command
        ocr[n] := SPIGet;                                                       // If the CMD8 was accepted, the card's functions get enhanced: now ACMD41 and CMD58 are supported - see below.
      EndFor;
      If (ocr[2] = $01) AND (ocr[3] = $AA) Then
        tmr := 12000;
        While ((tmr > 0) AND (MMCSendCmd(ACMD41, $40,0,0,0, $FF) <> 0)) Do      // Keep sending the ACMD41 for ~ 1 sec., until the card is no longer busy (initialization complete). $40 indicates, that we support SDHC cards.
          tmr := tmr -1;
        EndWhile;
        If (tmr <> 0) AND (MMCSendCMD(CMD58, 0, 0, 0, 0, $FF) = 0) Then         // The Card was ready, so we send CMD58
          For n := 0 To 3 Do                                                    // and we read back the Operation Condition Register (OCR) with the Card Capacity status (CCS)
            ocr[n] := SPIGet;
          EndFor;
          If ((ocr[0] and $40) = $40) Then                                      // Check the CCS - if it is set to 1
            CardType := CT_SD2 OR CT_BLOCK;                                     // we have a SDHC card - here we address the data in complete Sectors (i.e. Blocks)
          Else
            CardType := CT_SD2;                                                 // Else we have a SD V2 card
          EndIf;
        EndIf;
      EndIf;
    Else                                                                        // The CMD8 was not supported, so
{$IFDEF MMC_SDV1_Support}
      If (MMCSendCmd(ACMD41, 0, 0, 0, 0, $FF) <= 1) Then                        // we check, if the card accepts ACMD41
        CardType := CT_SD1;                                                     // Yes, so we have a SD Version 1 card
        cmd := ACMD41;                                                          // prepare to send ACMD41 for ~ 2 sec.
      Else                                                                      // The card did not support ACMD41
        CardType := CT_MMC;                                                     // so we may have a MMC card here - it uses CMD1 instead of ACMD41
        cmd := CMD1;                                                            // prepare to send CMD1 for ~ 2 sec.
      EndIf;

      tmr := 25000;
      While ((tmr > 0) AND (MMCSendCmd(cmd, 0, 0, 0, 0, $FF) <> 0)) Do          // we keep sending the prepared command
        tmr := tmr -1;
      EndWhile;

      If (tmr = 0) OR (MMCSendCmd(CMD16, 0, 0, 2, 0, $FF) <> 0) Then            // CMD16: Set the BlockLength for writing or reading to 512 Bytes
        CardType := 0;                                                          // If we had a timeout previously or CMD16 was not accepted, we have NO valid card.
      EndIf;
{$ENDIF}
    EndIf;
  INCL(MMC_CS);                                                                 // deactivate the CS signal
End;


{===============================================================================
  MMCReadSector
===============================================================================}
FUNCTION MMCReadSector(sec : LONGWORD) : BOOLEAN;                               // Read in one complete sector (512 bytes) from the MMC
VAR
  i : Word;
  res : BYTE;
  bc : word;
  rc : byte;
BEGIN
  sec := sec + PTI_StartSector;                                                 // If we have a MBR on the card, we need to add the Start-Sector offset of the current partition to all values
  If ((CardType AND CT_BLOCK) = 0) Then                                         // We have a standard SD (and NOT a SDHC) card: Addressing not in Sectors but Bytes
     sec := sec SHL 9;                                                          // SHL 9 is the same as * 512, which is the Sector Size.. SD-Cards do the adressing in Bytes, not in Sectors like the SDHC cards
  EndIf;

  SPCR := SPIREADCLOCK;                                                         // Set the reading speed
  INCL(SPSR, 0);
  res := MMCSendCmd(CMD17,                                                      // Send the Read_Single_Block Command with the address of the block
                    HI(HIWORD(sec)), LO(HIWORD(sec)),
                    HI(LOWORD(sec)), LO(LOWORD(sec)), $FF);

  IF res <> 0 THEN                                                              // Error, return
    INCL(MMC_CS);
    RETURN(FALSE);
  ENDIF;

  bc := 30000;
  Repeat
    rc := SPIGet;
    bc := bc -1;
  Until ((rc <> $FF) OR (bc = 0));

  If (rc = $FE) Then                                                            // The block is ready for reading
     i := 0;
     WHILE i <= BPB_BytesPerSec - 1 DO                                          // Read in the complete sector
        buffer[i] := SPIGet;                                                    // into our buffer
        Inc(i);
     ENDWHILE;
  EndIf;
  SPIGet;
  SPIGet;
  INCL(MMC_CS);                                                                 // deactivate the CS signal
  SPIGet;
  RETURN (TRUE);
END MMCReadSector;


{===============================================================================
  FATInit
===============================================================================}
FUNCTION FATInit : BOOLEAN;                                                     // Initialize the Card, get the CardType and initialize the FAT system
BEGIN
  PTI_StartSector := 0;                                                         // We start on the absolute first sector of the card
  CardInit;                                                                     // We initialize the Card, which sets the global variable "CardType"
  IF (CardType = 0) Or NOT MMCReadSector(0) THEN                                // If we have no valid CardType or the Reading of the first sector fails
    RETURN (FALSE);                                                             // We return with the failure
  ENDIF;
                                                                                // The first sector tells us the Begin of the first Partition
  CopyBlock(@buffer[$1BE + $08], @PTI_StartSector,4);                           // Copy the Longword from the buffer into the Variable PTI_StartSector
  IF NOT MMCReadSector(0) THEN                                                  // Now we read the Sector 0 again, this time with the correct offset of PTI_StartSector
    RETURN (FALSE);                                                             // Again, if this fails, we leave
  ENDIF;
                                                                                // Now we read in all the important information
  FATver              := tFAT16;                                                // Get the FAT type - we assume, it is FAT16
  FAT_RootDirSecs     := 32;                                                    // Get the number of sectors of the Root Directory - For FAT16 this is 32
  BPB_SecPerCluster   := buffer[13];                                            // Get the number of Sectors in one Cluster
  BPB_NumFATs         := buffer[16];                                            // Get the number of FAT tables
  CopyBlock(@buffer[22], @BPB_FATSz, 2);                                        // Get the number of sectors of one FAT table
  BPB_FATSz           := BPB_FATSz And $0000FFFF;                               // BPB_FATSz is a LONGWORD: Do not forget to delete the HiWord of that LongWord variable!
  CopyBlock(@buffer[14], @BPB_RsvdSecCnt, 2);                                   // Get the number of Reserved Sectors

  If BPB_FatSz = 0 Then                                                         // If the Result of the FAT Size was 0, it is NOT a FAT16. It is FAT32
     CopyBlock(@buffer[36], @BPB_FATSz, 4);                                     // Then the FAT Size is located in these 4 Bytes
     FAT_RootDirSecs := 0;                                                      // And there is no seperate Root Directory. (FAT32 contains the Root in the same area as all the other Data, starting at Cluster 2)
     FATver := tFAT32;                                                          // So we have FAT32
  EndIf;

  FAT_RootDir   := LONGWORD(BPB_RsvdSecCnt) + BPB_FatSz * LONGWORD(BPB_NumFATs);// Calculate the first sector of the Root Directory: it begins after the reserved sectors and after the FATs

  FAT_StartData := FAT_RootDir + LongWord(FAT_RootDirSecs);                     // Calculate the first sector of the File/Directory/Data area: for FAT16 it begins after the Root Directory.
                                                                                // For FAT32 it is the same sector as FAT_RootDir
  RETURN (TRUE);                                                                // All is done and OK.
END FATInit;


{===============================================================================
  FindNextCluster
  Sucht den nchsten Cluster
===============================================================================}
FUNCTION FindNextCluster : BOOLEAN;                                             // Searches the FAT for the next cluster number of the current cluster
VAR
  ByteOffset  : LONGWORD;
  SectorOffset: LONGWORD;
BEGIN
  ByteOffset := LONGWORD(NextCluster) * 2 * LongWord(FATVer);                   // Calculate the Byte Address of the current cluster in the FAT.
                                                                                // FAT16 has 2 Byte FAT Entries (FAT16 can not have more than 65.524 clusters!), FAT32 has 4 Byte FAT Entries.
  SectorOffset := (ByteOffset DIV BPB_BytesPerSec) + LONGWORD(BPB_RsvdSecCnt);  // Calculate the absolute Sector address of the FAT, that contains the current Cluster's entry
  IF MMCReadSector(SectorOffset) THEN                                           // and read that sector into the buffer
     CopyBlock(@buffer[LOWORD(ByteOffset AND 511)],@NextCluster,2*Word(FATVer));// Within this buffer of 512 bytes, calculate the address of the FAT Entry, it can be only 0..511.
                                                                                // Then copy it to the Target, i. e. NextCluster. Depending on the FATVer we have a Word (FAT16) or a LongWord (FAT32)
     If (FATVer = tFAT32) Then                                                  // If we have FAT32:
        Return( Not (NextCluster >= $0FFFFFF8));                                // If the Cluster reads $0F FF FF F8 or higher, there is no following cluster
     Else                                                                       // If we have FAT16:
        NextCluster := NextCluster AND $0000FFFF;                               // Do not forget to delete the HiWord of that Longword Variable!
        Return( Not (NextCluster >= $FFF8));                                    // If the Cluster reads $FF F8 or higher, there is no following cluster
     EndIf;
  ENDIF;
  RETURN (FALSE);
END FindNextCluster;


{===============================================================================
  FindFirstCluster
===============================================================================}
FUNCTION FindFirstCluster : BOOLEAN;                                            // Searches for the FAT entry of our firmware update file (e.g. CPU.BIN) in the Root Directory and returns the first Cluster of the file.
VAR
  sec : Byte;
  d,e : BYTE;
  p   : tDirPtr;
  CurrSec : LongWord;
BEGIN
   sec := 0;                                                                    //
   CurrSec := FAT_RootDir;                                                      // We start with the first Sector of the Root Directory (FAT16), resp. the first Sector of the FAT table (FAT32)
   NextCluster := 2;                                                            // FAT32 writes the Root Dir in the DataArea, starting in Cluster 2
   WHILE sec < (32 * Byte(FATVer)) DO                                           // FAT16 has a fixed Root Dir area of 32 Sectors. However, for FAT32 we must read up to 64 Sectors per Cluster
      MMCReadSector(CurrSec + LONGWORD(sec));                                   // We read the current Root Dir sector in
      p := @Dir[1];                                                             // and point to the first record (which is 32 bytes long). So, one Sector (512 bytes) has 16 records (32 bytes): 16 x 32 = 512
      d := 1;                                                                   // d is a counter for the 16 directory entries within one sector
      WHILE d <= 16 DO                                                          // we check the FAT entries one by one, running through the full sector.
         WITH p^ DO                                                             // the pointer is pointing to the Beginning of a 32 Byte Directory record
            e := 1;                                                             //
            while e < 12 do                                                     // we compare each character of the file name and extension with our update file name ("CPU.BIN" = "CPU     BIN")
               If NAME_EXT[e] <> UPDNAME[e] Then Break; EndIf;                  // As soon as we have any difference, we leave the comaparison loop
               Inc(e);                                                          // otherwise we go to the next character
            endwhile;
            If e = 12 Then                                                      // If we finished the comparison without any difference, we have found the file!
               FirstCluster := LongWord(ClusterLo)+(LongWord(ClusterHi) SHL 16);// Get the first Cluster of the File from the Root Directory entry
               NextCluster  := FirstCluster;                                    //
               RETURN (TRUE);                                                   // Fine, we found it!
            ENDIF;
         ENDWITH;
         p++;                                                                   // go the the next FAT entry
         Inc(d);                                                                // Increase the directory entry counter
      ENDWHILE;
      Inc(sec);                                                                 // The file's directory entry was not in this sector. So, increase the Root Dir Sector counter
      If (FATVer = tFAT32) Then                                                 // With FAT32 we have to check more:
         If Sec = BPB_SecPerCluster Then                                        // if we have reached the maximum number of Sectors within a Cluster, we must search the next cluster of the Root Directory
            If FindNextCluster Then                                             // Here we do it. If we found it, we must calculate the new Sector number from the Cluster number, since this routine uses Sector addresses
               CurrSec := LONGWORD(BPB_SecPerCluster) * (NextCluster-2) + FAT_RootDir;
               Sec := 0;                                                        // Do not forget to reset the running sector counter to 0
            Else
               Break;                                                           // we have reached the end of the Root Directory, so we leave the loop
            EndIf;
         EndIf;
      EndIf;
   ENDWHILE;
   RETURN(FALSE);                                                               // We could not find the file, so we give up.
END FindFirstCluster;


{===============================================================================
  FlashBlock
===============================================================================}
Procedure FlashBlock(BlockAdr : Byte);                                          // This Procedure will load one FlashBlock (256 Bytes = 128 Words) into the Processors RAM FLASH buffer and Flash it
Begin
   FlashErasePage;                                                              // Erase the current FLASH-Page (This is a 256 Byte Block)
   FlashCopyR2F(@flash[BlockAdr]);                                              // Copy from the read sector buffer, beginning at BlockAdr, 256 Bytes into the Processor's RAM FLASH buffer
   FlashProgPage;                                                               // Write the RAM FLASH buffer into the FLASH
   Inc(FLASH_ADDR,$100);                                                        // Increase the Pointer by 256 bytes
   If FLASH_ADDR = 0 Then Inc(FLASH_PAGE); EndIf;                               // If it has flown over, we must increase the FLASH_PAGE: careful!! This is not a 256 byte page, but a 65.535 Byte area
                                                                                // A 256 KByte Processor like the ATMEGA 2561 has 4 Pages to address the full flash memory
   Inc(WordsFlashed,$80);                                                       // Increase the count of the written WORDS - this helps to find out, when we are ready with flashing.
End;


{===============================================================================
  GetCluster
===============================================================================}
FUNCTION GetCluster : BOOLEAN;                                                  // This Procedure will read one Cluster of the Update File and Flash it
VAR
  i   : BYTE;
  sec : LONGWORD;
BEGIN
  sec := LONGWORD(BPB_SecPerCluster) * (NextCluster-2) + FAT_StartData;         // Calculate the Sector number of the current Cluster
  i := 1;
  WHILE i <= BPB_SecPerCluster DO                                               // While we have not reached the number of Sectors per Cluster
    IF MMCReadSector(sec) AND (WordsFlashed <= ((Boot_Start Shr 1) - 256)) THEN // Read the Sector in and IF we have not yet written the last Page (256 bytes) before the Bootloader itself
       FlashBlock(0);                                                           // Flash the first  256 Bytes of the Sector and..
       FlashBlock(128);                                                         // Flash the second 256 Bytes of the Sector (Here addressed in WORDS, thus only 128)
    ELSE
      RETURN (FALSE);                                                           // Else: Some Error happened, leave
    ENDIF;
    Inc(Sec);                                                                   // Go to the next Sector
    Inc(i);                                                                     // Increase the Counter
  ENDWHILE;
  RETURN (TRUE);                                                                // Finished.
END GetCluster;


{===============================================================================
  UpdateFirmware
  Fhrt den Update durch
===============================================================================}
PROCEDURE UpdateFirmware;                                                       // This Procedure will run the complete UPDATE process
BEGIN
   If FindFirstCluster Then                                                     // Search the first Cluster of the new Firmware File. If it was found,
      FlashInitPage(ADDR(Application_Start));                                   // Initialize the Flashing Routines and set the Pointer, where to start. (This setzt FLASH_ADDR and FLASH_PAGE)
      WordsFlashed := 0;                                                        // We have not flashed any words so far
      GetCluster;                                                               // Get and Flash the first Cluster
      WHILE FindNextCluster DO                                                  // While there are more Clusters in the File:
        GetCluster;                                                             // Get the next Cluster and Flash it.
      ENDWHILE;
   EndIf;                                                                       // The END. That was all. No, it was no easy job.
END UpdateFirmware;

INITIALIZATION
END MMCLoader16_32.

